// flop-io.c
//
// Flop is Copyright (c) 1995,2010 by Jason Summers
// This program is distributed under the terms of the GNU General
// Public License, version 3 or higher.

#define WINVER         0x0500
#define _WIN32_WINNT   0x0500
#define _WIN32_WINDOWS 0x0500

#include <windows.h>
#include <tchar.h>
#include <commdlg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "flop.h"
#include "resource.h"
#include <strsafe.h>

void fl_mb_to_w(const char *a, WCHAR *w, int w_len, int cp)
{
	int r;
	w[0] = '\0';
	r=MultiByteToWideChar(cp, 0, a, -1, w, w_len);
	w[w_len-1] = '\0';
}

void fl_w_to_mb(const WCHAR *w, char *a, int a_len, int cp)
{
	a[0] = '\0';
	WideCharToMultiByte(cp, 0, w, -1, a, a_len, NULL, NULL);
	a[a_len-1] = '\0';
}

void fl_a_to_w(const char *a, WCHAR *w, int w_len)
{
	fl_mb_to_w(a,w,w_len,CP_ACP);
}

void fl_w_to_a(const WCHAR *w, char *a, int a_len)
{
	fl_w_to_mb(w,a,a_len,CP_ACP);
}

void fl_utf8_to_w(const char *a, WCHAR *w, int w_len)
{
	fl_mb_to_w(a,w,w_len,CP_UTF8);
}

void fl_w_to_utf8(const WCHAR *w, char *a, int a_len)
{
	fl_w_to_mb(w,a,a_len,CP_UTF8);
}

// Allocates ctx->filebuf, containing the contents of the file.
// Caller must free.
// On failure, ctx->filebuf will be set to NULL.
static int fl_FileToMem(struct Ctx *ctx, const TCHAR *filename)
{
	HANDLE fh=INVALID_HANDLE_VALUE;
	DWORD bytesread=0;
	int retval;

	ctx->filebuf=NULL;
	ctx->filebuf_pos=0;

	fh=CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		goto done;
	}

	ctx->filebuf_size = GetFileSize(fh,NULL);
	if(ctx->filebuf_size<1 || ctx->filebuf_size>1000000) {
		goto done;
	}

	ctx->filebuf = calloc(ctx->filebuf_size,1);
	if(!ctx->filebuf) goto done;

	ReadFile(fh,ctx->filebuf,ctx->filebuf_size,&bytesread,NULL);
	if(bytesread!=ctx->filebuf_size) goto done;

	retval=1;

done:
	if(fh!=INVALID_HANDLE_VALUE) CloseHandle(fh);
	if(!retval && ctx->filebuf) {
		free(ctx->filebuf);
		ctx->filebuf=NULL;
	}
	return retval;
}

// Returns 0 on failure.
static int fl_ReadLine(struct Ctx *ctx, TCHAR *buf, int buf_len)
{
	int bufpos=0;
	TCHAR c;

	if(ctx->filebuf_pos>=(int)ctx->filebuf_size) return 0; // End of "file".

	while(1) {
		if(ctx->filebuf_pos>=(int)ctx->filebuf_size) goto done;
		if(bufpos>=buf_len-1) goto done;
		c=ctx->filebuf[ctx->filebuf_pos];
		buf[bufpos] = c;
		if(c=='\r' || c=='\n') buf[bufpos]='\0'; // Don't return EOL chars.

		bufpos++;
		ctx->filebuf_pos++;
		if(c=='\n') goto done;
	}

done:
	buf[bufpos]='\0';
	return 1;
}

void fl_WriteFileF(HANDLE fh, const TCHAR *fmt, ...)
{
	va_list ap;
	TCHAR buf[1024];
#ifdef UNICODE
	char bufA[1024];
#endif
	int len;
	DWORD byteswritten;

	va_start(ap, fmt);
	StringCbVPrintf(buf,sizeof(buf),fmt,ap);
	va_end(ap);

#ifdef UNICODE
	fl_w_to_a(buf,bufA,1024);
	len=strlen(bufA);
	WriteFile(fh,bufA,len,&byteswritten,NULL);
#else
	len=strlen(buf);
	WriteFile(fh,buf,len,&byteswritten,NULL);
#endif
}

// fbuf_len is the max number of characters (bytes) to read from fbuf,
// not including NUL byte(s). A NUL byte will terminate the read, but
// there does not have to be a NUL byte.
// s_len includes room for the NUL terminator. 's' will always be NUL
// terminated by this function,
static void fl_ReadFileBufToString(const char *fbuf1, int fbuf1_len,
		 TCHAR *s, int s_len)
{
	char fbuf[101];

	s[0]='\0';
	if(fbuf1_len>100) return;
	CopyMemory(fbuf,fbuf1,fbuf1_len); // make a copy which is NUL terminated.
	fbuf[fbuf1_len]='\0';

#ifdef UNICODE
	fl_utf8_to_w(fbuf,s,s_len);
#else
	StringCchCopy(s,s_len,fbuf);
#endif
}

// fbuf_len includes room for a terminating NUL. A NUL terminator shouldn't
// really be required here, but the file format apparently expects it.
// TODO: Should pay more attention to character encodings.

static void fl_WriteStringToFileBuf(const TCHAR *s, char *fbuf, int fbuf_len)
{
	ZeroMemory(fbuf,fbuf_len);
#ifdef UNICODE
	fl_w_to_utf8(s, fbuf, fbuf_len);
#else
	StringCchCopy(fbuf,fbuf_len,s);
#endif
}

static TCHAR *GetNameForLoad(struct Ctx *ctx, HWND HWindow, const TCHAR *szTemp,
			const TCHAR *title, const TCHAR *defext)
{
	OPENFILENAME ofnTemp;
	static TCHAR FileName[MAX_PATH];

	ZeroMemory(&ofnTemp,sizeof(OPENFILENAME));

	ofnTemp.lStructSize = sizeof( OPENFILENAME );
	ofnTemp.hwndOwner = HWindow;
	//ofnTemp.hInstance = 0;
	ofnTemp.lpstrFilter = szTemp;
	//ofnTemp.lpstrCustomFilter = NULL;
	//ofnTemp.nMaxCustFilter = 0;
	ofnTemp.nFilterIndex = 1;
	ofnTemp.lpstrFile = FileName;
	FileName[0]='\0';
	ofnTemp.nMaxFile = sizeof(FileName)/sizeof(TCHAR);
	//ofnTemp.lpstrFileTitle = NULL;
	//ofnTemp.nMaxFileTitle = 0;
	//ofnTemp.lpstrInitialDir = NULL;
	ofnTemp.lpstrTitle = title;
	ofnTemp.Flags = OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
	//ofnTemp.nFileOffset = 0;
	//ofnTemp.nFileExtension = 0;
	ofnTemp.lpstrDefExt = defext;
	//ofnTemp.lCustData = 0;
	//ofnTemp.lpfnHook = NULL;
	//ofnTemp.lpTemplateName = NULL;

	if(GetOpenFileName( &ofnTemp ) != TRUE) {
		DWORD Errval=CommDlgExtendedError();
		if(Errval!=0) {   // 0 value means user selected Cancel
			StringCchPrintf(ctx->buffer,100,_T("File error #%u selecting load file"),(unsigned int)Errval);
			Error(ctx,ctx->buffer);
		}
		return NULL;
	}
	return FileName;
}

static TCHAR *GetNameForSave(struct Ctx *ctx, HWND HWindow, const TCHAR *szTemp,
							const TCHAR *title, const TCHAR *defext, int warn_overwrite)
{
	OPENFILENAME ofnTemp;
	static TCHAR FileName[MAX_PATH];

	ZeroMemory(&ofnTemp,sizeof(OPENFILENAME));
	ofnTemp.lStructSize = sizeof( OPENFILENAME );
	ofnTemp.hwndOwner = HWindow;
	//ofnTemp.hInstance = 0;
	ofnTemp.lpstrFilter = szTemp;
	//ofnTemp.lpstrCustomFilter = NULL;
	//ofnTemp.nMaxCustFilter = 0;
	ofnTemp.nFilterIndex = 1;
	ofnTemp.lpstrFile = FileName;
	FileName[0]='\0';
	ofnTemp.nMaxFile = sizeof(FileName)/sizeof(TCHAR);
	//ofnTemp.lpstrFileTitle = NULL;
	//ofnTemp.nMaxFileTitle = 0;
	//ofnTemp.lpstrInitialDir = NULL;
	ofnTemp.lpstrTitle = title;
	ofnTemp.Flags = OFN_HIDEREADONLY | OFN_NOREADONLYRETURN;
	if(warn_overwrite) ofnTemp.Flags |= OFN_OVERWRITEPROMPT;
	//ofnTemp.nFileOffset = 0;
	//ofnTemp.nFileExtension = 0;
	ofnTemp.lpstrDefExt = defext;
	//ofnTemp.lCustData = 0;
	//ofnTemp.lpfnHook = NULL;
	//ofnTemp.lpTemplateName = NULL;

	if(GetSaveFileName( &ofnTemp ) != TRUE) {
		DWORD Errval=CommDlgExtendedError();
		if(Errval!=0) {   // 0 value means user selected Cancel
			StringCchPrintf(ctx->buffer,100,_T("File error #%u selecting save file"),(unsigned int)Errval);
			Error(ctx,ctx->buffer);
		}
		return NULL;
	}
	return FileName;
}

int LoadLevel(struct Ctx *ctx, struct Game *g, int levelnum)
{
	int i,j;
	int retval=0;
	unsigned char buf[LEVELSIZE];
	WORD s;
	HANDLE fh=INVALID_HANDLE_VALUE;
	DWORD retdw;
	BOOL retb;
	DWORD bytesread;

	fh=CreateFile(_T("flop.dat"),GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		Error(ctx,_T("File error: Unable to open flop.dat"));
		goto done;
	}
	fl_InitGame(g);

	retdw=SetFilePointer(fh,(LONG)((levelnum-1)*LEVELSIZE),NULL,FILE_BEGIN);
	if(retdw==INVALID_SET_FILE_POINTER) {
		if(GetLastError()!=NO_ERROR) {
			Error(ctx,_T("File error: Error searching flop.dat"));
			goto done;
		}
	}

	retb=ReadFile(fh,(void*)buf,LEVELSIZE,&bytesread,NULL);
	if(!retb || bytesread!=LEVELSIZE) {
		Error(ctx,_T("File error: Error reading flop.dat"));
		goto done;
	}

	s=buf[0]+256*buf[1];
	g->bestscore= (s>32767)? -1 : (int)s;
	// TODO: Don't hardcode 21, 65

	fl_ReadFileBufToString(&buf[2],20, g->bestname,21);
	fl_ReadFileBufToString(&buf[23],20, g->levelname,21);
	fl_ReadFileBufToString(&buf[44],64, g->message,65);

	for(j=0;j<12;j++)
		for(i=0;i<12;i++)
			g->field[i+1][j+1]=buf[109+j*12+i];
	g->pos.x=(int)buf[253];
	g->pos.y=(int)buf[254];
	g->needed=(int)buf[255];

	retval=1;

done:
	if(fh!=INVALID_HANDLE_VALUE) {
		CloseHandle(fh);
	}
	return retval;
}

// b is the level to save
int SaveLevel(struct Ctx *ctx, struct Game *g,int b,int scoreonly)
{
	int i,j;
	int retval=0;
	HANDLE fh=INVALID_HANDLE_VALUE;
	DWORD byteswritten;
	unsigned char buf[LEVELSIZE];
	int num;
	DWORD retdw;
	BOOL retb;

	num=LEVELSIZE;
	if(scoreonly) num=23;

	if(b<=20 && !scoreonly && !ctx->editall) {
		Error(ctx,_T("You cannot change the first 20 levels"));
		return 0;
	}

	if(scoreonly) {
		if(g->bestscore<0) {
			buf[0]=buf[1]=255;
		}
		else {
			buf[0]=(unsigned char)(g->bestscore % 256);
			buf[1]=(unsigned char)(g->bestscore / 256);
		}
		fl_WriteStringToFileBuf(g->bestname,&buf[2],FL_MAXNAMELEN+1);
		num=2+FL_MAXNAMELEN+1;     // this is getting ugly
	}
	else {
		buf[0]=buf[1]=255;
		fl_WriteStringToFileBuf(_T(""),&buf[2],FL_MAXNAMELEN+1);
		fl_WriteStringToFileBuf(g->levelname,&buf[23],FL_MAXNAMELEN+1);
		fl_WriteStringToFileBuf(g->message,&buf[44],FL_MAXMESSAGELEN+1);

		for(j=0;j<12;j++)
			for(i=0;i<12;i++)
				buf[109+j*12+i]=(unsigned char)g->field[i+1][j+1];
		buf[253]=(unsigned char)g->pos.x;
		buf[254]=(unsigned char)g->pos.y;
		buf[255]=(unsigned char)g->needed;
	}

	fh = CreateFile(_T("flop.dat"),GENERIC_WRITE,0,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		Error(ctx,_T("File error: Unable to open flop.dat for writing"));
		goto done;
	}

	retdw=SetFilePointer(fh,(LONG)((b-1)*LEVELSIZE),NULL,FILE_BEGIN);

	if(retdw==INVALID_SET_FILE_POINTER) {
		Error(ctx,_T("File error: Error searching flop.dat"));
		goto done;
	}

	retb=WriteFile(fh,buf,num,&byteswritten,NULL);
	if(!retb || (int)byteswritten!=num) {
		Error(ctx,_T("File error: Error writing flop.dat"));
		goto done;
	}

	retval=1;

done:
	if(fh != INVALID_HANDLE_VALUE) CloseHandle(fh);
	return retval;
}

struct fl_importdlg_struct {
	int selectindex;
	struct Ctx *ctx;
};

static void OnImportINITDIALOG(struct fl_importdlg_struct *fli, HWND hWnd)
{
	TCHAR buf[FL_MAXNAMELEN+2];

	while(1) {
		if(!fl_ReadLine(fli->ctx,buf,FL_MAXNAMELEN+2))
			return;  // end of file reached

		buf[FL_MAXNAMELEN+1]=0;

		if(buf[0]=='>') {
			SendDlgItemMessage(hWnd,IDC_LISTBOX, LB_ADDSTRING,0,
							(DWORD)(&buf[1]) );
		}
	}
}

static void OnImportOK(struct fl_importdlg_struct *fli, HWND hDlg)
{
	fli->selectindex=(int)SendDlgItemMessage(hDlg,IDC_LISTBOX,LB_GETCURSEL,0,0L);
	if(fli->selectindex==LB_ERR) fli->selectindex= -1;
}

static INT_PTR CALLBACK ImportDlg(HWND hDlg,UINT msg,WPARAM wParam,LPARAM lParam)
{
	WORD id;
	struct fl_importdlg_struct *fli = NULL;

	if(msg==WM_INITDIALOG) {
		SetWindowLongPtr(hDlg,DWLP_USER,lParam);
		fli=(struct fl_importdlg_struct*)lParam;
	}
	else {
		fli=(struct fl_importdlg_struct*)GetWindowLongPtr(hDlg,DWLP_USER);
	}
	if(!fli) return FALSE;

	switch(msg) {
	case WM_INITDIALOG:
		OnImportINITDIALOG(fli,hDlg);
		return TRUE;

	case WM_COMMAND:
		id=LOWORD(wParam);

		switch(id) {
		case IDC_LISTBOX:
			if(HIWORD(wParam)==LBN_DBLCLK) {
				// User double-clicked on a listbox item.
				OnImportOK(fli,hDlg);
				EndDialog(hDlg, TRUE);
				return TRUE;
			}
			break;

		case IDOK:
			OnImportOK(fli,hDlg);
			EndDialog(hDlg, TRUE);
			return TRUE;

		case IDCANCEL:
			fli->selectindex= -1;
			EndDialog(hDlg, TRUE);
			return TRUE;
		}
		break;
	}

	return FALSE;		/* Didn't process a message */
}

int Import(struct Ctx *ctx, struct Game *newgame)
{
	int i,j;
	TCHAR *filename;
#ifdef UNICODE
	TCHAR buf[256];
	TCHAR x;
#else
	unsigned char buf[256];
	unsigned char x;
#endif
	int retval=0;
	struct fl_importdlg_struct fli;

	ctx->filebuf=NULL;

	//	 strcpy(textbuf,"");
	//	 if(!GetText(_T("File to import from:"),0)) return 0;
	filename=GetNameForLoad(ctx,ctx->hwndMain,_T("*.lev\0*.lev\0*.*\0*.*\0"),_T("Import level"),
												_T("lev"));
	if(!filename) goto done;

	fl_FileToMem(ctx,filename);
	if(!ctx->filebuf) {
		Error(ctx,_T("File error: Can't open file"));
		goto done;
	}

	fli.ctx = ctx;
	fli.selectindex = -1;
	DialogBoxParam(ctx->hInst,_T("ImportBox"),ctx->hwndMain,ImportDlg,(LPARAM)&fli);

	if(fli.selectindex<0) { goto done; }  /* canceled by user */

	ctx->filebuf_pos=0; // Seek back to beginning of "file"

	for(i=0;i<=fli.selectindex;i++) {
		do {
			if(!fl_ReadLine(ctx,buf,128))
				return 0;  // end of file reached too soon
		} while(buf[0]!='>');

	}

	fl_InitGame(newgame);

	StringCchCopy(newgame->levelname,FL_MAXNAMELEN+1,&buf[1]);
	fl_ReadLine(ctx,buf,128);
	StringCchCopy(newgame->message,FL_MAXMESSAGELEN+1,buf);
	fl_ReadLine(ctx,buf,128);
	newgame->needed = _tstoi(buf);
	for(j=1;j<=TY;j++) {
		fl_ReadLine(ctx,buf,128);
		for(i=1;i<=TX;i++) {
			newgame->field[i][j]=0;
			x=buf[(i-1)*2];
			if(x>='@' && x<='O') {
				newgame->field[i][j] |= ((x&15)<<3)|I_TILE;
			}
			switch(buf[(i-1)*2+1]) {
			case 's': newgame->pos.x=i; newgame->pos.y=j; break;
			case 'r': newgame->field[i][j]|=I_RING;      break;
			case 'x': newgame->field[i][j]|=I_X;         break;
			case 'd': newgame->field[i][j]|=I_DOOR;      break;
			}
		}
	}
	fl_NewGame(ctx);
	UpdateAll(ctx);
	retval=1;
done:
	if(ctx->filebuf) { free(ctx->filebuf); ctx->filebuf=NULL; }
	return retval;
}

int Export(struct Ctx *ctx, struct Game *newgame)
{
	int i,j,x;
	TCHAR c1,c2, *filename;
	HANDLE fh=INVALID_HANDLE_VALUE;
	int retval=0;

	filename=GetNameForSave(ctx,ctx->hwndMain,_T("*.lev\0*.lev\0"),_T("Export level"),_T("lev"),0);
	if(!filename) return 0;

	fh=CreateFile(filename,GENERIC_WRITE,0,NULL,OPEN_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		goto done;
	}

	// Seek to end of file, so we will append to it if it already exists.
	SetFilePointer(fh,0,NULL,FILE_END);

	fl_WriteFileF(fh,_T(">%s\r\n"),newgame->levelname);
	fl_WriteFileF(fh,_T("%s\r\n"),newgame->message);
	fl_WriteFileF(fh,_T("%d\r\n"),newgame->needed);

	for(j=1;j<=TY;j++) {
		for(i=1;i<=TX;i++) {
			x=newgame->field[i][j];
			if(x&I_TILE) c1=(char)(64+( (x&I_ALLH) >> 3 ));
			else c1='.';
			if(i==newgame->pos.x && j==newgame->pos.y) c2='s';
			else if(x&I_RING) c2='r';
				else if(x&I_X) c2='x';
					else if(x&I_DOOR) c2='d';
						else c2='.';
			fl_WriteFileF(fh,_T("%c%c"),c1,c2);
		}
		fl_WriteFileF(fh,_T("\r\n"));
	}
	retval=1;

done:
	if(fh!=INVALID_HANDLE_VALUE) CloseHandle(fh);
	return retval;
}

// m points to the position in the move buffer to start with
// returns the number of moves added to the buffer
// max is the maximum number of moves to load
// returns number of moves loaded
int LoadMoves(struct Ctx *ctx, unsigned char *m,int m_size)
{
	int i,ret=0;
	TCHAR *filename;
	DWORD bytesread;
	HANDLE fh;

	filename=GetNameForLoad(ctx,ctx->hwndMain,_T("*.mv\0*.mv\0*.*\0*.*\0"),_T("Load moves"),_T("mv"));
	if(!filename) return 0;

	fh=CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		Error(ctx,_T("File error: Unable to open file"));
	}
	else {
		if(ReadFile(fh,m,m_size,&bytesread,NULL)) {
			ret = (int)bytesread;
		}
		else {
			Error(ctx,_T("File error: Unable to read file"));
			ret=0;
		}
		CloseHandle(fh);
	}
	for(i=0;i<ret;i++) m[i] &= 3;   // strip meaningless bits
	return ret;
}

// n is number of moves to save   m points to move buffer
// returns 1 on success, 0 on failure
int SaveMoves(struct Ctx *ctx, unsigned char *m,int num)
{
	int i,ret=0;
	TCHAR *filename;
	HANDLE fh;
	DWORD byteswritten;
	int retval=0;

	if(num<1) {
		Error(ctx,_T("Internal: No moves to save"));
		return 0;
	}
	for(i=0;i<num;i++) m[i] |= 48;       // makes moves human-readable

	filename=GetNameForSave(ctx,ctx->hwndMain,_T("*.mv\0*.mv\0*.*\0*.*\0"),_T("Save moves"),_T("mv"),1);
	if(!filename) goto done;

	fh=CreateFile(filename,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if(fh==INVALID_HANDLE_VALUE) {
		Error(ctx,_T("File error: Unable to open file"));
		goto done;
	}

	if(!WriteFile(fh,m,num,&byteswritten,NULL)) {
		CloseHandle(fh);
		Error(ctx,_T("File error: Error writing file"));
		goto done;
	}

	CloseHandle(fh);

	if(ret) {
		StringCchPrintf(ctx->textbuf,256,_T("%d move%s saved"),(int)num,
								( ((int)num) ==1)?_T(""):_T("s")  );
		MessageBox(ctx->hwndMain,ctx->textbuf,_T(""),MB_OK|MB_ICONINFORMATION);
	}

done:
	 for(i=0;i<num;i++) m[i] &= 3;      // restore moves

	 return ret;
}
